/* 
  BE n1 - Base de donnes
  Implmentation de la classe BaseDonnees
  2001-10-05, Mathieu Goutelle
*/

#include <string.h>
#include <stdio.h>
#include "BD.h"

// Constructeurs
//   - nom_fic contient le nom du fichier contenant les donnes
//   - nom_fic_ini contient le nom du fichier contenant les donnes de 
//       remplissage et de dbordement
BaseDonnees::BaseDonnees() {
  int i;
  FILE* ptr_fic;

  nom_fic= new char[11];
  strcpy(nom_fic, "defaut.txt");
  nom_fic_ini = new char[11];
  strcpy(nom_fic_ini, "defaut.ini");
  nb_art = 10;

  // Initialisation des infos de remplissage et de dbordement  
  for (i=0; i<nb_paq; i++) {
    debord[i] = 0;
    jauge[i] = 0;
  };

  // On remplit le fichier d'espace 
  ptr_fic = fopen(nom_fic, "w+b");
  for(i=0;i<(nb_paq*taille_paq);i++)
    fprintf(ptr_fic, "%c",  32);
  fclose(ptr_fic);
}

BaseDonnees::BaseDonnees(char *nom, int n) {
  int i;
  FILE *ptr_fic;

  nom_fic = new char[strlen(nom)+5];
  strcpy(nom_fic, nom);
  strcat(nom_fic, ".txt");
  nom_fic_ini = new char[strlen(nom)+5];
  strcpy(nom_fic_ini, nom);
  strcat(nom_fic_ini, ".ini");
  nb_art = n;
  
  if ((ptr_fic = fopen(nom_fic, "r+b")) == NULL) {
    // le fichier n'existe pas : on le cre
    ptr_fic = fopen(nom_fic, "w+b");
    for (i=0; i<(nb_paq*taille_paq); i++)
      fprintf(ptr_fic, "%c", 32);
    for (i=0; i<nb_paq; i++) {
      debord[i] = 0;
      jauge[i] = 0;
    };
  }
  else {
    // le fichier existe : on charge les donnes
    fclose(ptr_fic);
    ptr_fic = fopen(nom_fic_ini, "r+b");
    char buf[1];
    for (i=0; i<nb_paq; i++) {
      fread(buf, sizeof(char), 1, ptr_fic);
      debord[i] = buf[0] - '0';
    };
    for (i=0; i<nb_paq; i++) {
      fread(buf, sizeof(char), 1, ptr_fic);
      jauge[i] = buf[0] - '0';
    };
  };
  fclose(ptr_fic);
}

BaseDonnees::BaseDonnees(const BaseDonnees &bd) {
  nom_fic = new char[strlen(bd.nom_fic)+1];
  strcpy(nom_fic, bd.nom_fic);
  nb_art = bd.nb_art;

  int i;
  for (i=0; i<nb_paq; i++) {
    debord[i] = bd.debord[i];
    jauge[i] = bd.jauge[i];
  };
}

BaseDonnees::~BaseDonnees() {
  FILE *ptr_fic;
  int i;
  ptr_fic = fopen(nom_fic_ini, "w+b");

  // Sauvegarde des informations de dbordement
  for (i=0; i<nb_paq; i++)
    fprintf(ptr_fic, "%c", '0'+debord[i]);
  // Sauvegarde des informations de remplissage
  for (i=0; i<nb_paq; i++)
    fprintf(ptr_fic, "%c", '0'+jauge[i]);
  fclose(ptr_fic);

//  delete[] nom_fic;
//  delete[] nom_fic_ini;
}

// Compare deux articles : le caractre '*' est considr comme un caractre
// gnrique (joker) pour la recherche multi-critres
int BaseDonnees::Compare_art(char art1[65], char art2[65]) const {
  int result = 1;
  int i = 0;

  while (result && (i<64)) {
    result &= (art1[i] == art2[i] || art1[i] == '*');
    i++;
  };

  return result;
}

// Recherche dans num_paq (qui contient aussi le numro de l'article initial dans le paquet) 
// l'article spcifi. Renvoie le numro de l'article s'il a t trouv, -1 sinon.
int BaseDonnees::Recherche_ds_paq(int num_paq, char article[65], FILE *ptr_fic) const {
  char buffer[65];
  int i, trouve;
  int i_ini = num_paq%nb_art; // numro de l'article initial
  int num_paq_cur = num_paq/nb_art; // numro du paquet initial

  fseek(ptr_fic, num_paq_cur*taille_paq + i_ini*64, SEEK_SET);
  do {
    // On parcourt les paquets
    num_paq_cur %= nb_paq; // pour permettre de chainer le dernier article au premier
    i = i_ini;
    do { 
      // on cherche jusqu' ce que l'on trouve ou que l'on sorte du paquet
      fread(buffer, sizeof(char), 64, ptr_fic);
      trouve = Compare_art(article, buffer);
      i++;
    } while (!trouve && i<nb_art);
    i_ini = 0;
  // Si le paquet courant a dbord et que l'on a pas trouv, on continue.
  // On s'arrte si l'on a fait un tour complet (tous les paquets ont dbord)
  } while (a_deborde(num_paq_cur++) && !trouve && (num_paq_cur)%nb_paq != (num_paq/nb_art));

  if (!trouve)
    return -1;
  else
    return ((num_paq_cur-1)%nb_paq)*nb_art + i-1;
}

// Accs aux attributs
char* BaseDonnees::Get_nom_fic() const { 
  char *result = new char[strlen(nom_fic)+1];
  strcpy(result, nom_fic);
  return result;
}

int BaseDonnees::Get_nb_art() const { return nb_art; }

int BaseDonnees::a_deborde(int nb) const {
  return debord[nb];
}

int BaseDonnees::est_vide(int nb) const {
  return (jauge[nb] == 0);
}

int BaseDonnees::est_plein(int nb) const {
  return (jauge[nb] == nb_art);
}

// MAJ des attributs
void BaseDonnees::Set_nom_fic(char* n) {   
  delete[] nom_fic;  
  nom_fic = new char[strlen(n)+1];
  strcpy(nom_fic,n);
}

void BaseDonnees::Set_nb_art(int n) { nb_art = n; }

void BaseDonnees::deborde(int n) { debord[n] = 1; }

void BaseDonnees::remplit(int n) { jauge[n]++; }
void BaseDonnees::vide(int n) { jauge[n]--; }

// Oprations sur la BD
float BaseDonnees::tx_debord() const {
  int i, result=0;
  for (i=0; i<nb_paq; i++)
    result += debord[i];

  return (float)result/(float)nb_paq;
}

float BaseDonnees::tx_rempl() const {
  int i, result=0;
  for (i=0; i<nb_paq; i++)
    result += jauge[i];

  return (float)result/(float)(nb_paq*nb_art);
}

int BaseDonnees::Insertion(const Employe &Emp) {
  FILE *ptr_fic;
  int num_paq = Emp.hachage();
  int pos_ds_fic;

  // On recherche le premier paquet non plein. On teste si l'on a pas fait
  // un tour complet. Si le paquet est plein et qu'il n'a pas encore dbord,
  // on le marque comme ayant dbord.
  while (est_plein(num_paq) && ((num_paq+1)%nb_paq != Emp.hachage())) {
    if (!a_deborde(num_paq)) deborde(num_paq);
    num_paq++; num_paq %= nb_paq;
  };

  ptr_fic = fopen(nom_fic, "r+b");
  pos_ds_fic = Recherche_ds_paq(num_paq*nb_art, article_vide, ptr_fic);

  if (pos_ds_fic == -1) {
    return 0; // la BD est pleine
  }
  else {
    // On crit l'article  la position prcdemment calcule
    fseek(ptr_fic, pos_ds_fic*64, SEEK_SET);
    fwrite(Emp.article(), sizeof(char), 64, ptr_fic);
    remplit(pos_ds_fic/nb_art);
  };
  fclose(ptr_fic);
  return 1;
}

int BaseDonnees::Suppression(const Employe &Emp) {
  FILE *ptr_fic;
  int num_paq = Emp.hachage();
  int pos_ds_fic;

  ptr_fic = fopen(nom_fic, "r+b");
  pos_ds_fic = Recherche_ds_paq(num_paq*nb_art, Emp.article(), ptr_fic);
  if (pos_ds_fic == -1) {
    return 0; // l'lment est absent de la base
  }
  else {
    // On efface l'article
    fseek(ptr_fic, pos_ds_fic*64, SEEK_SET);
    fwrite(article_vide, sizeof(char), 64, ptr_fic);
    vide(pos_ds_fic/nb_art);
  };
  fclose(ptr_fic);
  return 1;
}

int BaseDonnees::Modification(const Employe &ori, const Employe &nouv) {
  Suppression(ori);
  Insertion(nouv);
}

// Construction de l'article avec les informations fournies par l'utilisateur.
// Si l'utilisateur n'a pas renseign un champ, il est remplac par des '*'.
char* construit_article(unsigned int N,
			unsigned int S, char opS, 
			char* D) {

  char art[65] = article_vide;
  int i, q;

  if (N) {
    for(i=0, q=10000; i<5; i++, q /= 10)
      art[i] = '0'+(N/q)%10;
  }
  else {
    for(i=0; i<5; i++)
      art[i] = '*';
  };

  if (S && opS == '=') {
    for(i=5, q=10000; i<10; i++, q /= 10)
      art[i] = '0'+(S/q)%10;
  }
  else {
    for(i=5; i<10; i++)
      art[i] = '*';
  };

  if (strlen(D)) {
    for(i=0; i<strlen(D); i++) art[i+10] = D[i];
  }
  else
    for(i=10; i<37; i++) art[i] = '*';

  for(i=37; i<64; i++) art[i] = '*';

  return art;
}

int BaseDonnees::Recherche(unsigned int N, 
			   unsigned int S, char opS, 
			   char* D) const {
  FILE *ptr_fic;
  int num, sal, dep;
  int tmp_num, tmp_sal, tmp_dep;
  int pos_ds_fic;
  char buffer[64];
  Employe *emp_temp;

  char *article = new char[65];
  // Construction de l'article  rechercher
  if (S==0) opS = '=';
  strcpy(article, construit_article(N, S, opS, D));

  ptr_fic = fopen(nom_fic,"r+b");
  // On passe en revue toutes les cls. Si l'utilisateur a spcifi le champ correspondant,
  // on sort immdiatement de la boucle en tenant compte des infos fournies (cf. instruction
  // 'break' en fin de boucle).
  for (num=0; num<16; num++) {
    for (sal=0; sal<4; sal++) {
      for (dep=0; dep<8; dep++) {
	// Si l'utilisateur a fourni des infos, on en tient compte. Sinon, on prend
	// le compteur de boucle pour crer la cla de hachage
	if (N) pos_ds_fic = 32*(N%16); else pos_ds_fic = 32*num;
	if (S) pos_ds_fic += 8*(S%4); else pos_ds_fic += 8*sal;
	if (strlen(D)) pos_ds_fic += strlen(D)%8; else pos_ds_fic += dep;
	pos_ds_fic *= nb_art;

	// On saute les paquets vides et ceux ne contenant pas l'article
 	while (!est_vide(pos_ds_fic/nb_art) && 
	       (pos_ds_fic = Recherche_ds_paq(pos_ds_fic, article, ptr_fic)) != -1) {
	  fseek(ptr_fic, pos_ds_fic*64, SEEK_SET);
	  fread(buffer, sizeof(char), 64, ptr_fic);

	  if (strcmp(buffer, article_vide) != 0) {
	  //if (Compare_art(buffer, article_vide) == 1) {
	    // Pour une raison que nous ignorons, 'strcmp' ne fonctionne pas parfois,
	    // mais est bien plus rapide que notre propre fonction. D'o la double comparaison...
	    if (Compare_art(buffer, article_vide) == 1) break;
	    // On affiche si l'employ satisfait au critre sur le salaire
	    emp_temp = new Employe(buffer);
	    switch (opS) {
	    case '=': 
	      cout<<*emp_temp<<'\n';
	      break;
	    case '<':
	      if (emp_temp->Get_salaire() < S) cout<<*emp_temp<<'\n';
	      break;
	    case '>':
	      if (emp_temp->Get_salaire() > S)	cout<<*emp_temp<<'\n';
	      break;
	    default:
	      cout<<"Oprateur inconnu.\n";
	      return 0;
	    };
	    // On poursuit la recherche dans le mme paquet mais  partir de l'article suivant.
	    pos_ds_fic += 1;
	  };
	};
	if (strlen(D)) break;
      };
      if (S && opS=='=') break;
    };
    if (N) break;
  };
  fclose(ptr_fic);

  return 1;
}
